import numpy as np

from .kernels import gaussian


def estimate_bandwidth(x_train, kernel='gaussian'):
    """
    Bandwidth estimation.

    Parameters
    ----------
    x_train : ndarray of shape (m, n)
        Input.
    kernel : {'uniform', 'gaussian', 'epanechnikov', 'cauchy'}, default='gaussian'
        Kernel name.

    Returns
    -------
    bandwidth : array_like of shape (n,)
        Bandwidth (smoothing) parameter.
    """
    m_train = x_train.shape[0]
    std_x = np.std(x_train, axis=0, ddof=1)
    if m_train == 1:
        std_x = np.std(x_train, axis=0)
    std_x[std_x == 0] = 1

    if kernel == 'gaussian':
        W, U = 1 / (2 * np.sqrt(np.pi)), 1
    elif kernel == 'uniform':
        W, U = 1 / 2, 1 / 3
    elif kernel == 'cauchy':
        W, U = 5 / (4 * np.pi), 1
    elif kernel == 'epanechnikov':
        W, U = 0.6, 0.2
    else:
        raise ValueError(f'invalid kernel: {kernel}')
    WU2 = W / (U ** 2)
    Z = 3 / (8 * np.sqrt(np.pi))

    bandwidth = (WU2 / (Z * m_train)) ** (1 / 5) * std_x
    return bandwidth


class Kde():
    def fit(self, x_train, weights_train=None, bandwidth=None):
        self.m_train, self.n = x_train.shape
        self.x_train = np.copy(x_train)

        if weights_train is None:
            self.weights_train = np.full(self.m_train, 1 / self.m_train)
        else:
            self.weights_train = np.copy(weights_train)
            self.weights_train = self.weights_train / self.weights_train.sum()

        if bandwidth is None:
            self.bandwidth = estimate_bandwidth(self.x_train)
        elif bandwidth.any() <= 0:
            raise ValueError(f'bandwidth needs to be greater than zero or None, got {bandwidth}')
        else:
            self.bandwidth = np.copy(bandwidth)

        self.s = np.ones(self.m_train)
        return self

    def score_samples(self, x_test):
        scores = 1 / (np.prod(self.bandwidth)) * np.sum((self.weights_train / (self.s ** self.n))[:, None] * np.prod(
            gaussian((x_test - self.x_train[:, None]) / (self.bandwidth * self.s[:, None, None])), axis=2), axis=0)
        return scores



import setuptools

with open('VERSION', 'r') as f:
    version = f.read().strip()

with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    install_requires = f.read().split('\n')

setuptools.setup(
    name='kdelearn',
    version=version,
    author='Krystian Franus',
    author_email='krystian.franus@gmail.com',
    description='Short description of ml methods based on kernel density estimation',
    long_description=long_description,
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License'
    ],
    python_requires='>=3.6',
)



import numpy as np


def uniform(x):
    """
    Uniform kernel function.

    Parameters
    ----------
    x : :obj:`ndarray`
        Function argument.

    Returns
    -------
    ndarray
        Values of the function for the given arguments.
    """
    return np.where(np.abs(x) <= 1, 0.5, 0.0)


def gaussian(x):
    """
    Gaussian kernel function.

    Parameters
    ----------
    x : ndarray
        Function argument.

    Returns
    -------
    ndarray
        Values of the function for the given arguments.
    """
    return np.exp(-x ** 2 / 2) / np.sqrt(2 * np.pi)


def epanechnikov(x):
    """
    Epanechnikov kernel function.

    Parameters
    ----------
    x : ndarray
        Function argument.

    Returns
    -------
    ndarray
        Values of the function for the given arguments.
    """
    return np.where(np.abs(x) <= 1, 3 / 4 * (1 - x ** 2), 0.0)


def cauchy(x):
    """
    Cauchy kernel function.

    Parameters
    ----------
    x : ndarray
        Function argument.

    Returns
    -------
    ndarray
        Values of the function for the given arguments.
    """
    return 2 / (np.pi * (x ** 2 + 1) ** 2)
